/*
 *
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>

#include "utils/dlist.h"

#include "tdm_common.h"
#include "tdm_object.h"
#include "tdm_log.h"

#include "tdm_device.h"
#include "tdm_port.h"


struct tdm_device *tdm_object_to_device(struct tdm_object *obj)
{
	if (!obj || !tdm_object_is_device(obj))
		return NULL;

	return (struct tdm_device *)obj;
}

struct tdm_module *tdm_device_get_module(struct tdm_device *dev)
{
	if (!dev || !tdm_object_is_device(dev))
		return NULL;

	return (struct tdm_module *)tdm_object_owner(dev);
}

struct tdm_device *tdm_device_new(const char *name, const struct tdm_device_ops *ops)
{
	struct tdm_device *dev = NULL;

	if (!name || !name[0] || !ops)
		return NULL;

	dev = (struct tdm_device *)tdm_object_new(name, sizeof(*dev));
	if (!dev)
		return NULL;

	tdm_object_set_type(dev, TDM_OBJECT_DEVICE);

	dev->ops = (struct tdm_device_ops *)ops;

	return dev;
}

tdm_status_t tdm_device_destroy(struct tdm_device *dev)
{
	if (!dev || !tdm_object_is_device(dev))
		return TDM_STATUS_GENERR;

	/* TODO: check if port is still in use */
	if (dev->nr_ports) {
		tdm_error("%s:device still has ports\n", tdm_object_name(dev));
		return TDM_STATUS_GENERR;
	}

	if (dev->ops->device_close) {
		tdm_status_t retval;
		if ((retval = dev->ops->device_close(dev)) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to close device\n", tdm_object_name(dev));
			return retval;
		}
	} else {
		tdm_debug("%s: no device close callback\n", tdm_object_name(dev));
	}

	tdm_object_destroy((struct tdm_object *)dev);
	return TDM_STATUS_SUCCESS;
}

tdm_status_t tdm_device_add_port(struct tdm_device *device, struct tdm_port *port)
{
	struct tdm_module *owner = NULL;
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!device || !tdm_object_is_device(device))
		return TDM_STATUS_GENERR;

	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	owner = tdm_device_get_module(device);
	if (!owner) {
		tdm_error("%s: orphaned device", tdm_object_name(device));
		return TDM_STATUS_GENERR;
	}


	/* TODO: lock device */

	dlist_insert_tail(&device->ports, &port->port_list);
	device->nr_ports++;

	port->device = device;

	tdm_object_set_owner(port, owner);

	/* TODO: unlock device */

	/* */
	if (device->ops->device_add_port) {
		if ((retval = device->ops->device_add_port(device, port)) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to add port \"%s\"\n", tdm_object_name(device), tdm_object_name(port));
			goto error;
		}
	} else {
		tdm_debug("%s: no remove port callback\n", tdm_object_name(device));
	}

	return TDM_STATUS_SUCCESS;

error:
	dlist_remove(&port->port_list);
	device->nr_ports--;
	port->device = NULL;

	return retval;
}

int tdm_device_has_port(struct tdm_device *dev, struct tdm_port *port)
{
	if (!dev || !tdm_object_is_device(dev))
		return TDM_FALSE;

	if (!port || !tdm_object_is_port(port))
		return TDM_FALSE;

	return (dlist_get_head(&dev->ports) == dlist_get_head(&port->port_list));
}

tdm_status_t tdm_device_remove_port(struct tdm_device *device, struct tdm_port *port)
{
	struct tdm_module *owner = NULL;

	if (!device || !tdm_object_is_device(device))
		return TDM_STATUS_GENERR;

	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	if (!tdm_device_has_port(device, port)) {
		tdm_error("%s: not a member of device \"%s\"\n", tdm_object_name(port), tdm_object_name(device));
		return TDM_STATUS_GENERR;
	}

	/* */
	if (device->ops->device_remove_port) {
		tdm_status_t retval;
		if ((retval = device->ops->device_remove_port(device, port)) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to remove port \"%s\"\n", tdm_object_name(device), tdm_object_name(port));
			return retval;
		}
	} else {
		tdm_debug("%s: no remove port callback\n", tdm_object_name(device));
	}

	/* TODO: lock device */

	dlist_remove(&port->port_list);
	device->nr_ports--;
	port->device = NULL;

	/* TODO: unlock device */

	return TDM_STATUS_SUCCESS;
}
