/*
 *
 *
 */

#define TDM_FALSE	0
#define	TDM_TRUE	1

/********************************************************
 * Object
 ********************************************************/

typedef enum {
	TDM_OBJECT_NONE = 0,
	TDM_OBJECT_MODULE,
	TDM_OBJECT_DEVICE,
	TDM_OBJECT_PORT,
	TDM_OBJECT_PROTOCOL,
	TDM_OBJECT_SPAN,
	TDM_OBJECT_SPANGROUP,
	TDM_OBJECT_CHANNEL,
	TDM_OBJECT_CALL,

	TDM_OBJECT_MAX	/* always last! */
} tdm_object_type_t;

/**
 * @note	Always first member of struct!
 */
struct tdm_object {
	char *name;
	struct tdm_object *owner;
	tdm_object_type_t type;

	unsigned long get_cnt, put_cnt;	/*!< reference counting */
};

#define tdm_object_is(x, t)	\
	((struct tdm_object *)(x)->type == (t))

#define tdm_object_is_module(x)	\
	tdm_object_is(x, TDM_OBJECT_MODULE)

#define tdm_object_is_device(x)	\
	tdm_object_is(x, TDM_OBJECT_DEVICE)

#define tdm_object_is_port(x)	\
	tdm_object_is(x, TDM_OBJECT_PORT)

#define tdm_object_is_span(x)	\
	tdm_object_is(x, TDM_OBJECT_SPAN)

#define tdm_object_is_spangroup(x)	\
	tdm_object_is(x, TDM_OBJECT_SPANGROUP)

#define tdm_object_is_channel(x)	\
	tdm_object_is(x, TDM_OBJECT_CHANNEL)

#define tdm_object_is_protocol(x)	\
	tdm_object_is(x, TDM_OBJECT_PROTOCOL)

#define tdm_object_name(x)	\
	((struct tdm_object *)(x)->name))

#define tdm_object_type(x)	\
	((struct tdm_object *)(x)->type))

#define tdm_object_owner(x)	\
	((struct tdm_object *)(x)->owner))

#define tdm_object_set_type(x, t)	\
	((struct tdm_object *)(x)->type = (t))

#define tdm_object_set_owner(x, o)	\
	(struct tdm_object *)(x)->owner = (struct tdm_object *)(o)

/* not sure if we want this */
#define tdm_object_use_count(x)	\
	((struct tdm_object *)(x)->get_cnt - (struct tdm_object *)(x)->put_cnt)

#define tdm_object_in_use(x)	\
	(tdm_object_use_count(x) != 0)

#define tdm_object_put(x)	\
	({ (struct tdm_object *)(x)->put_cnt++; (x); })

#define tdm_object_get(x)	\
	({ (struct tdm_object *)(x)->get_cnt++; (x); })


struct tdm_object *tdm_object_new(const char *name, const size_t size)
{
	struct tdm_object *obj = NULL;

	if (!name || !name[0] || size < sizeof(struct tdm_object))
		return NULL;

	obj = calloc(size);
	if (!obj)
		return NULL;

	obj->type = TDM_OBJECT_NONE;

	obj->name = strdup(name);
	if (!obj->name) {
		free(obj);
		return NULL;
	}
	return obj;
}

/**
 * Rename object, use with care!
 */
tdm_status_t tdm_object_rename(struct tdm_object *obj, const char *name)
{
	char *old_name = NULL;
	char *new_name = NULL;

	if (!obj)
		return TDM_STATUS_GENERR;

	old_name = obj->name;

	new_name = strdup(name);
	if (!new_name)
		return TDM_STATUS_GENERR;

	obj->name = new_name;

	free(old_name);
	return TDM_STATUS_SUCCESS;
}

void tdm_object_destroy(struct tdm_object *obj)
{
	if (!obj)
		return;

	free(obj->name);
	free(obj);
}


/***********************************************************
 * Sig: Span
 ***********************************************************/


typedef enum {
	TDM_SPAN_STATE_NONE = 0,
	TDM_SPAN_STATE_INVALID	/* always last */
} tdm_span_state_t;

struct tdm_span {
	struct tdm_object obj;

	struct tdm_span_ops *ops;

	tdm_span_state_t    state;

	unsigned int        nr_channels;
	struct dlist_head   channels;

	/* grouping */
	struct dlist_head   group_span_list;
};

struct tdm_spangroup *tdm_span_get_spangroup(struct tdm_span *span)
{
	struct tdm_spangroup *group = NULL;
	struct dlist_head *entry;

	if (!span || !tdm_object_is_span(span))
		return NULL;

	entry = dlist_get_head(&span->group_span_list);
	if (!entry) {
		tdm_error("%s: span not member of a group\n", tdm_object_name(span));
		return NULL;
	}

	group = dlist_entry(entry, struct tdm_spangroup, spans);
	if (!tdm_object_is_spangroup(group)) {
		tdm_error("%p: object is not a span group\n", group);
		return NULL;
	}

	return group;
}

struct tdm_span *tdm_object_to_span(struct tdm_object *obj)
{
	if (!obj || !tdm_object_is_span(obj))
		return NULL;

	return (struct tdm_span *)obj;
}

struct tdm_span *tdm_span_new(const char *name)
{
	struct tdm_span *span = NULL;

	span = (struct tdm_span *)tdm_object_new(name, sizeof(*span));
	if (!span)
		return NULL;

	tdm_object_set_type(span, TDM_OBJECT_SPAN);

	dlist_init(&span->channels);

	return span;
}

tdm_status_t tdm_span_add_channel(struct tdm_span *span, struct tdm_channel *chan)
{
	tdm_status_t retval = TDM_STATUS_SUCCES;

	if (!span || !tdm_object_is_span(span))
		return TDM_STATUS_GENERR;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	if (tdm_channel_has_span(chan)) {
		tdm_error("%s: already assigned to span \"%s\"\n", tdm_object_name(chan), tdm_object_name(tdm_channel_get_span(chan)));
		return TDM_STATUS_GENERR;
	}

	/* TODO: locking */

	dlist_append(&span->channels, &chan->span_channel_list);
	span->nr_channels++;
	chan->span = span;

	/* TODO: locking */

	/* let signalling module handle span add operation
	 * TODO: do this before, of after adding the channel to the span??
	 */
	if (span->ops->span_add_channel) {
		if ((retval = span->ops->span_add_channel(span, chan)) != TDM_STATUS_SUCCESS) {
			tdm_log_error("%s: failed to remove channel \"%s\"\n", tdm_object_name(span), tdm_object_name(chan));
			goto error;
		}
	} else {
		tdm_debug("%s: no add channel callback\n", tdm_object_name(span));
	}

	return TDM_STATUS_SUCCESS;

error:
	dlist_remove(&chan->span_channel_list);
	span->nr_channels--;
	chan->span = NULL;
	return retval;
}

int tdm_span_has_channel(struct tdm_span *span, struct tdm_channel *chan)
{
	if (!span || !tdm_object_is_span(span))
		return TDM_FALSE;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_FALSE;

	return (dlist_get_head(chan->span_channel_list) == dlist_get_head(&span->channels));
}

tdm_status_t tdm_span_remove_channel(struct tdm_span *span, struct tdm_channel *chan)
{
	if (!span || !tdm_object_is_span(span))
		return TDM_STATUS_GENERR;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	if (!tdm_span_has_channel(span, chan)) {
		tdm_error("%s: not a member of span \"%s\"\n", tdm_object_name(chan), tdm_object_name(span));
		return TDM_STATUS_GENERR;
	}

	/* TODO: more checks? e.g. channel in use? */

	/* see if signalling module allows removing the channel */
	if (span->ops->span_remove_channel) {
		tdm_status_t retval;
		if ((retval = span->ops->span_remove_channel(span, chan)) != TDM_STATUS_SUCCESS) {
			tdm_log_error("%s: failed to remove channel \"%s\"\n", tdm_object_name(span), tdm_object_name(chan));
			return retval;
		}
	} else {
		tdm_debug("%s: no remove channel callback\n", tdm_object_name(span));
	}

	/* TODO: locking */

	dlist_remove(&chan->span_channel_list);
	span->nr_channels--;
	chan->span = NULL;

	/* TODO: locking */

	return TDM_STATUS_SUCCESS;
}

/************************************************************
 * Span goups / Span group policy
 ***********************************************************/

struct tdm_policy_def {
	const char *name;
	const char *description;

	tdm_status_t (policy_init *)(struct tdm_policy *policy);
	tdm_status_t (policy_destroy *)(struct tdm_policy *policy);
	tdm_status_t (policy_configure *)(struct tdm_policy *policy);

	/* => struct tdm_spangroup_ops(?) */
	tdm_status_t (group_add_span *)(struct tdm_spangroup *group, struct tdm_span *span);
	tdm_status_t (group_remove_span *)(struct tdm_spangroup *group, struct tdm_span *span);
};

struct tdm_policy {
	struct tdm_object obj;

	struct tdm_policy_def *def;
};

static DLIST_HEAD(tdm_policy_list);

tdm_status_t tdm_module_register_policy(struct tdm_module *mod, const struct tdm_policy_def *def);
{
	struct tdm_policy *policy = NULL;

	policy = tdm_object_new(def->name, sizeof(*policy));
	if (!policy)
		return TDM_STATUS_GENERR;

	tdm_object_set_type(policy, TDM_OBJECT_POLICY);
	tdm_object_set_owner(policy, mod);


	/* TOOD: locking */

/*
	if (tdm_hash_insert(tdm_object_name(policy), policy) != TDM_STATUS_SUCCESS) {
		tdm_error("%s: failed to register protocol\n", name);
		goto error;
	}
 */

	dlist_append(&tdm_policy_list, &policy->policy_list);

	/* TOOD: locking */

	return TDM_STATUS_SUCCESS;

error:
	tdm_object_destroy(policy);
	return TDM_STATUS_GENERR;
}


struct tdm_spangroup {
	struct tdm_object obj;

	struct tdm_policy  *policy	/*!< policy for this group */

	unsigned int       nr_spans;
	struct dlist_head  spans;

	void *priv;			/*!< custom data for policy */
};

void tdm_spangroup_set_private(struct tdm_spangroup *group, void *data)
{
	if (!group || !tdm_object_is_spangroup(group))
		return;

	group->priv = data;
}

void *tdm_spangroup_get_private(struct tdm_spangroup *group)
{
	if (!group || !tdm_object_is_spangroup(group))
		return NULL;

	return group->priv;
}

struct tdm_spangroup *tdm_spangroup_new(const char *name)
{
	struct tdm_spangroup *group = NULL;

	group = tdm_object_new(name, sizeof(*group));
	if (!group)
		return NULL;

	tdm_object_set_type(group, TDM_OBJECT_SPANGROUP);

	dlist_init_head(&group->spans);

	return group;
}

tdm_status_t tdm_spangroup_add_span(struct tdm_spangroup *group, struct tdm_span *span)
{
	tdm_status_t retval = TDM_STATUS_SUCCES;

	if (!group || !tdm_object_is_spangroup(group))
		return TDM_STATUS_GENERR;

	if (!span || !tdm_object_is_span(span))
		return TDM_STATUS_GENERR;

	/* TODO: check if span is already a member of a/this group */

	/* TODO: locking */

	dlist_append(&group->spans, &span->group_span_list);
	group->nr_spans++;
	span->group = group;

	/* TODO: locking */

	/* */
	if (group->policy->group_add_span) {
		if ((retval = group->policy->group_add_span(group, span)) != TDM_STATUS_SUCCESS) {
			tdm_log_error("%s: failed to add span \"%s\"\n", tdm_object_name(group), tdm_object_name(span));
			goto error;
		}
	} else {
		tdm_debug("%s: no add span callback\n", tdm_object_name(group));
	}

	return TDM_STATUS_SUCCESS;

error:
	dlist_remove(&span->group_span_list);
	group->nr_spans--;
	span->group = NULL;
	return retval;
}

int tdm_spangroup_has_span(struct tdm_spangroup *group, struct tdm_span *span)
{
	if (!group || !tdm_object_is_spangroup(group))
		return TDM_FALSE;

	if (!span || !tdm_object_is_span(span))
		return TDM_FALSE;

	return (dlist_get_head(&span->group_span_list) == dlist_get_head(&group->spans));
}

tdm_status_t tdm_spangroup_remove_span(struct tdm_spangroup *group, struct tdm_span *span)
{
	if (!group || !tdm_object_is_spangroup(group))
		return TDM_STATUS_GENERR;

	if (!span || !tdm_object_is_span(span))
		return TDM_STATUS_GENERR;

	if (!tdm_spangroup_has_span(group, span)) {
		tdm_error("%s: not a member of spangroup \"%s\"\n", tdm_object_name(span), tdm_object_name(group));
		return TDM_STATUS_GENERR;
	}

	/* */
	if (group->policy->group_remove_span) {
		tdm_status_t retval;
		if ((retval = group->policy->group_remove_span(group, span)) != TDM_STATUS_SUCCESS) {
			tdm_log_error("%s: failed to remove span \"%s\"\n", tdm_object_name(group), tdm_object_name(span));
			return retval;
		}
	} else {
		tdm_debug("%s: no remove span callback\n", tdm_object_name(group));
	}

	/* TODO: locking */

	dlist_remove(&span->group_span_list);
	group->nr_spans--;
	span->group = NULL;

	/* TODO: locking */

	return TDM_STATUS_SUCCESS;
}


/************************************************************
 * I/O: Channel
 ***********************************************************/

struct tdm_channel {
	struct tdm_object obj;

	struct tdm_port        *port;
	struct tdm_span        *span;
	struct tdm_channel_ops *ops;

	struct dlist_head      span_channel_list;	/* => span->channels */
	struct dlist_head      port_channel_list;	/* => port->channels */
};

struct tdm_channel *tdm_channel_new(const char *name)
{
	struct tdm_channel *chan = NULL;

	chan = (struct tdm_channel *)tdm_object_new(name, sizeof(*chan));
	if (!chan)
		return NULL;

	tdm_object_set_type(span, TDM_OBJECT_CHANNEL);

	dlist_init(&chan->span_channel_list);
	dlist_init(&chan->port_channel_list);

	return chan;
}

struct tdm_module *tdm_channel_get_module(struct tdm_channel *chan)
{
	if (!chan || !tdm_object_is_channel(chan))
		return NULL;

	return (struct tdm_module *)tdm_object_owner(chan);
}

void tdm_channel_set_owner(struct tdm_channel *chan, struct tdm_module *owner)
{
	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	if (!owner || !tdm_object_is_module(owner))
		return TDM_STATUS_GENERR;

	tdm_object_set_owner(chan, owner);
}

struct tdm_channel *tdm_object_to_channel(struct tdm_object *obj)
{
	if (!obj || !tdm_object_is_channel(obj))
		return NULL;

	return (struct tdm_channel *)obj;
}

struct tdm_span *tdm_channel_get_span(struct tdm_channel *chan)
{
	struct tdm_span   *span = NULL;
	struct dlist_head *head;

	if (!chan || !tdm_object_is_channel(chan))
		return NULL;
#if 0
	head = dlist_get_head(&chan->span_channel_list);
	if (!head) {
		tdm_error("%s: orphaned or unassigned channel\n", tdm_object_name(chan));
		return NULL;
	}

	span = dlist_entry(head, struct tdm_span, channel_list);
	if (!tdm_object_is_span(span)) {
		tdm_error("%p: object is not a span object\n", span);
		return NULL;
	}
#else
	span = chan->span;
	if (!tdm_object_is_span(span)) {
		tdm_error("%p: object is not a span object\n", span);
		return NULL;
	}
#endif
	return span;
}

/* bad naming galore */
int tdm_channel_has_span(struct tdm_channel *chan)
{
	if (!chan || !tdm_object_is_channel(chan))
		return 0;
#if 0
	return !!(dlist_get_head(&chan->span_channel_list));
#else
	return !!(chan->span);
#endif
}

/****************************************************************
 * I/O: Device
 ****************************************************************/

struct tdm_device {
	struct tdm_object obj;

	struct tdm_device_ops *ops;

	unsigned int      nr_ports;
	struct dlist_head ports;
};

struct tdm_device *tdm_object_to_device(struct tdm_object *obj)
{
	if (!obj || !tdm_object_is_device(obj))
		return NULL;

	return (struct tdm_device *)obj;
}

tdm_status_t tdm_device_add_port(struct tdm_device *device, struct tdm_port *port)
{
	struct tdm_module *owner = NULL;
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!device || !tdm_object_is_device(device))
		return TDM_STATUS_GENERR;

	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	owner = tdm_device_get_module(device);
	if (!owner) {
		tdm_error("%s: orphaned device", tdm_object_name(device));
		return TDM_STATUS_GENERR;
	}


	/* TODO: lock device */

	dlist_append(&device->ports, &port->port_list);
	device->nr_ports++;

	port->device = device;

	tdm_object_set_owner(port, owner);

	/* TODO: unlock device */

	/* */
	if (device->ops->device_add_port) {
		if ((retval = device->ops->device_add_port(device, port)) != TDM_STATUS_SUCCESS) {
			tdm_log_error("%s: failed to add port \"%s\"\n", tdm_object_name(device), tdm_object_name(port));
			goto error;
		}
	} else {
		tdm_debug("%s: no remove port callback\n", tdm_object_name(device));
	}

	return TDM_STATUS_SUCCESS;

error:
	dlist_remove(&port->port_list);
	device->nr_ports--;
	port->device = NULL;

	return retval;
}

int tdm_device_has_port(struct tdm_device *dev, struct tdm_port *port)
{
	if (!dev || !tdm_object_is_device(dev))
		return TDM_FALSE;

	if (!port || !tdm_object_is_port(port))
		return TDM_FALSE;

	return (dlist_get_head(&dev->ports) == dlist_get_head(&port->port_list));
}

tdm_status_t tdm_device_remove_port(struct tdm_device *device, struct tdm_port *port)
{
	struct tdm_module *owner = NULL;

	if (!device || !tdm_object_is_device(device))
		return TDM_STATUS_GENERR;

	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	if (!tdm_device_has_port(device, port)) {
		tdm_error("%s: not a member of device \"%s\"\n", tdm_object_name(port), tdm_object_name(device));
		return TDM_STATUS_GENERR;
	}

	/* */
	if (device->ops->device_remove_port) {
		tdm_status_t retval;
		if ((retval = device->ops->device_remove_port(device, port)) != TDM_STATUS_SUCCESS) {
			tdm_log_error("%s: failed to remove port \"%s\"\n", tdm_object_name(device), tdm_object_name(port));
			return retval;
		}
	} else {
		tdm_debug("%s: no remove port callback\n", tdm_object_name(device));
	}

	/* TODO: lock device */

	dlist_remove(&device->port_list);
	devive->nr_ports--;
	port->device = NULL;

	/* TODO: unlock device */

	return TDM_STATUS_SUCCESS;
}

/****************************************************************
 * I/O: Port
 ****************************************************************/

struct tdm_port {
	struct tdm_object obj;

	struct tdm_device   *device;
	struct tdm_port_ops *ops;

	struct dlist_head   port_list;

	unsigned int        nr_channels;
	struct dlist_head   channels;
};

struct tdm_port *tdm_object_to_port(struct tdm_object *obj)
{
	if (!obj || !tdm_object_is_port(obj))
		return NULL;

	return (struct tdm_port *)obj;
}

struct tdm_module *tdm_port_get_module(struct tdm_port *port)
{
	if (!port || !tdm_object_is_port(port))
		return NULL;

	return (struct tdm_module *)tdm_object_owner(port);
}

struct tdm_port *tdm_port_new(const char *name, const struct tdm_port_ops *ops)
{
	struct tdm_port *port = NULL;

	if (!name || !name[0] || !ops)
		return NULL;

	port = tdm_object_new(name, sizeof(*port));
	if (!port)
		return NULL;

	tdm_object_set_type(port, TDM_OBJECT_PORT);

	port->ops = ops;

	return port;
}

tdm_status_t tdm_port_destroy(struct tdm_port *port)
{
	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	/* TODO: check if port is still in use */
	if (port->nr_channels) {
		tdm_error("%s: port still has channels\n", tdm_object_name(port));
		return TDM_STATUS_GENERR;
	}

	if (port->ops->port_destroy) {
		tdm_status_t retval;
		if ((retval = port->ops->port_destroy()) != TDM_STATUS_SUCCESS) {
			tdm_error("%s: failed to destroy port\n", tdm_object_name(port));
			return retval;
		}
	} else {
		tdm_debug("%s: no port destroy callback\n", tdm_object_name(port));
	}

	tdm_object_destroy(port);
	return TDM_STATUS_SUCCESS;
}

tdm_status_t tdm_port_add_channel(struct tdm_port *port, struct tdm_channel *chan)
{
	struct tdm_module *owner = NULL;
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	owner = tdm_port_get_module(port);
	if (!owner) {
		tdm_error("%s: orphaned port", tdm_object_name(port));
		return TDM_STATUS_GENERR;
	}

	/* TODO: lock port */

	dlist_append(&port->channels, &chan->port_channel_list);
	port->nr_channels++;

	chan->port = port;

	tdm_object_set_owner(chan, owner);

	/* TODO: unlock port */

	/* */
	if (port->ops->port_add_channel) {
		if ((retval = port->ops->port_add_channel(port, chan)) != TDM_STATUS_SUCCESS) {
			tdm_log_error("%s: failed to add channel \"%s\"\n", tdm_object_name(port), tdm_object_name(chan));
			goto error;
		}
	} else {
		tdm_debug("%s: no add channel callback\n", tdm_object_name(port));
	}

	return TDM_STATUS_SUCCESS;

error:
	dlist_remove();
	port->nr_channels--;
	chan->port = NULL;

	return retval;
}

int tdm_port_has_channel(struct tdm_port *port, struct tdm_channel *chan)
{
	if (!port || !tdm_object_is_port(port))
		return TDM_FALSE;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_FALSE;

	return (dlist_get_head(&chan->port_channel_list) == dlist_get_head(&port->channels));
}

tdm_status_t tdm_port_remove_channel(struct tdm_port *port, struct tdm_channel *chan)
{
	if (!port || !tdm_object_is_port(port))
		return TDM_STATUS_GENERR;

	if (!chan || !tdm_object_is_channel(chan))
		return TDM_STATUS_GENERR;

	if (!tdm_port_has_channel(port, channel)) {
		tdm_error("%s: not a member of port \"%s\"", tdm_object_name(chan), tdm_object_name(port));
		return TDM_STATUS_GENERR;
	}

	/* */
	if (port->ops->port_remove_channel) {
		tdm_status_t retval;
		if ((retval = port->ops->port_remove_channel(port, chan)) != TDM_STATUS_SUCCESS) {
			tdm_log_error("%s: failed to remove channel \"%s\"\n", tdm_object_name(port), tdm_object_name(chan));
			return retval;
		}
	} else {
		tdm_debug("%s: no remove channel callback\n", tdm_object_name(port));
	}


	/* TODO: lock port */

	dlist_remove(&chan->port_channel_list);
	port->nr_channels--;
	chan->port = NULL;

	/* TODO: unlock port */

	return TDM_STATUS_SUCCESS;
}


/****************************************************************
 * Module
 ****************************************************************/

/**
 * "Internal" module handle
 */
struct tdm_module {
	struct tdm_object obj;

	struct tdm_module_def *def;	/*!< module definition + ops */

	struct dlist_head     module_list;
};

/**
 * Definition of module
 */
struct tdm_module_def {
	const char *name;
	const char *description;

	/* module ops */
	tdm_status_t (module_load *)(struct tdm_module *mod);		/*!< invoked after loading */
	tdm_status_t (module_unload *)(struct tdm_module *mod);		/*!< invoked before unloading */
	tdm_status_t (module_probe *)(struct tdm_module *mod);		/*!< device auto-detection */
	tdm_status_t (module_configure *)(struct tdm_module *mod);	/*!< module configuration / manual device registraion */
};

/* add module to global list, after successfull load + init */
tdm_status_t tdm_module_add(struct tdm_module *mod)
{
	if (!mod || !tdm_object_is_module(mod))
		return TDM_STATUS_GENERR;

	/* TODO: lock tdm_module_list */

	dlist_append(&tdm_module_list, &mod->module_list);

	/* TODO: unlock tdm_module_list */
	return TDM_STATUS_SUCCESS;
}


/****************************************************************
 * Protocol / Span type
 ****************************************************************/

struct tdm_protocol_def {
	const char *name;
	const char *description;

	struct tdm_span_ops *span_ops;
};

/* aka "span type" */
struct tdm_protocol {
	struct tdm_object   obj;

//	struct tdm_span_ops *ops;
	struct tdm_protocol_def *def;

	struct dlist_head   protocol_list;
};

static DLIST_HEAD(tdm_protocol_list);

/**
 * signalling/mixed modules only: register new signalling type
 * @note	Usually only called from module_load
 */
#if 0
tdm_status_t tdm_module_register_protocol(struct tdm_module *mod, const char *name, const struct tdm_span_ops *ops)
{
	struct tdm_protocol *proto = NULL;

	if (!mod || !tdm_object_is_module(mod))
		return TDM_STATUS_GENERR;

	proto = (struct tdm_protocol *)tdm_object_new(name, sizeof(*proto));
	if (!proto)
		return TDM_STATUS_GENERR;

	tdm_object_set_type(proto, TDM_OBJECT_PROTOCOL);
	tdm_object_set_owner(proto, mod);

	proto->ops = ops;

	/* TODO: lock tdm_protocol_list */

/*	if (tdm_hash_insert(name, proto) != TDM_STATUS_SUCCESS) {
		tdm_error("%s: failed to register protocol\n", name);
		goto error;
	}
*/
	dlist_append(&tdm_protocol_list, &proto->protocol_list);

	/* TODO: unlock tdm_protocol_list */
	return TDM_STATUS_SUCCESS;

error:
	tdm_object_destroy(proto);
	return TDM_STATUS_GENERR;
}
#else
tdm_status_t tdm_module_register_protocol(struct tdm_module *mod, const struct tdm_protocol_def *def)
{
	struct tdm_protocol *proto = NULL;

	if (!mod || !tdm_object_is_module(mod) || !def)
		return TDM_STATUS_GENERR;

	proto = (struct tdm_protocol *)tdm_object_new(def->name, sizeof(*proto));
	if (!proto)
		return TDM_STATUS_GENERR;

	tdm_object_set_type(proto, TDM_OBJECT_PROTOCOL);
	tdm_object_set_owner(proto, mod);

	proto->def = def;

	/* TODO: lock tdm_protocol_list */

/*	if (tdm_hash_insert(name, proto) != TDM_STATUS_SUCCESS) {
		tdm_error("%s: failed to register protocol\n", name);
		goto error;
	}
*/
	dlist_append(&tdm_protocol_list, &proto->protocol_list);

	/* TODO: unlock tdm_protocol_list */
	return TDM_STATUS_SUCCESS;

error:
	tdm_object_destroy(proto);
	return TDM_STATUS_GENERR;
}
#endif

/**
 * @note	usually called only from module_unload callback
 */
tdm_status_t tdm_protocol_unregister(struct tdm_protocol *proto)
{
	if (!proto || !tdm_object_is_protocol(proto))
		return TDM_STATUS_GENERR;

	/* TODO: check if protocol is still in use */
	if (tdm_object_in_use(proto)) {
		tdm_error("%s: protocol still in use\n", tdm_object_name(proto));
		return TDM_STATUS_GENERR;
	}

	/* TODO: lock protocol_list */

//	tdm_hash_remove(tdm_object_name(proto));

	dlist_remove(&proto->protocol_list);

	/* TODO: unlock protocol_list */

	return TDM_STATUS_SUCCESS;
}

struct tdm_protocol *tdm_protocol_find(const char *name)
{
	struct dlist_entry *entry;

	dlist_for_each(&tdm_protocol_list, entry) {
		struct tdm_protocol *proto = dlist_entry(entry, struct tdm_protocol, protocol_list);

		if (!tdm_object_is_protocol(proto)) {
			tdm_error("%p: object in protocol list is not a protocol", proto);
			break;
		}

		if (!strcmp(tdm_object_name(proto), name))
			return proto;
	}
	return NULL;
}

/***************************************************************
 *
 ***************************************************************/

/**
 * @brief Used in modules to define callbacks
 */
struct tdm_span_ops {
	tdm_status_t (span_start *)(struct tdm_span *span);	/* call module's span start handler */
	tdm_status_t (span_stop *)(struct tdm_span *span);	/* call module's span stop handler */

	tdm_status_t (span_add_channel *)(struct tdm_span *span, struct tdm_channel *chan);
	tdm_status_t (span_remove_channel *)(struct tdm_span *span, struct tdm_channel *chan);
};

/**
 * @brief I/O Modules per-channel ops
 * @todo  rework for event based I/O scheme
 */
struct tdm_channel_ops {
	tdm_status_t (channel_open *)(struct tdm_channel *chan, const int flags);
	tdm_status_t (channel_read *)(struct tdm_channel *chan, char *buffer, size_t *size);		/* read data incoming on channel */
	tdm_status_t (channel_write *)(struct tdm_channel *chan, const char *buffer, size_t *size);	/* write data outgoing on channel */
	tdm_status_t (channel_close *)(struct tdm_channel *chan);
};

struct tdm_device_ops {
	tdm_status (device_open *)(struct tdm_device *dev);
	tdm_status (device_close *)(struct tdm_device *dev);

	tdm_status (device_add_port *)(struct tdm_device *dev, struct tdm_port *port);
	tdm_status (device_remove_port *)(struct tdm_device *dev, struct tdm_port *port);
};

struct tdm_port_ops {
	tdm_status (port_open *)(struct tdm_port *port);
	tdm_status (port_close *)(struct tdm_port *port);

	tdm_status (port_add_channel *)(struct tdm_port *port, struct tdm_channel *chan);
	tdm_status (port_remove_channel *)(struct tdm_port *port, struct tdm_channel *chan);
};


/******************************************************
 * Events (/ Messages / Notifications)
 ******************************************************/

typedef enum {
	TDM_EVENT_NONE = 0,
	TDM_EVENT_SPAN,
	TDM_EVENT_CHANNEL,
	TDM_EVENT_PORT,
	TDM_EVENT_DEVICE,
	TDM_EVENT_MAX			/*!< always last! */
} tdm_event_class_t;

struct tdm_event {
	int ev_class, ev_type;		/* e.g. TDM_EVENT_SPAN : TDM_SPAN_EVENT_READY */

	union {
		struct tdm_span    *span;
		struct tdm_channel *channel;
		struct tdm_device  *device;
		struct tdm_port    *port;
	} ev_obj;

	union {
		void *custom;
	} ev_data;

	void (ev_destroy *)(struct tdm_event *event);
};

struct tdm_event *tdm_event_new(const int ev_class)
{
	struct tdm_event *event = NULL;

	event = calloc(sizeof(*event));
	if (!event)
		return NULL;

	event->ev_class = ev_class;

	return event;
}

tdm_status_t tdm_event_queue(struct tdm_event *event)
{
	return TDM_STATUS_SUCCESS;
}

int tdm_event_class(const struct tdm_event *event)
{
	if (!event)
		return TDM_EVENT_NONE;

	return event->ev_class;
}

void tdm_event_set_class(struct tdm_event *event, const int ev_class)
{
	event->ev_class = ev_class;
}



typedef enum {
	TDM_SPAN_EVENT_NONE = 0,
	TDM_SPAN_EVENT_SPAN_READY,	/*!< Span started */
	TDM_SPAN_EVENT_SPAN_STOPPED,	/*!< Span stopped */
	TDM_SPAN_EVENT_MAX		/*!< always last! */
} tdm_span_event_t;


/**
 * callback to destroy span event
 */
static void tdm_span_destroy_event(struct tdm_event *event)
{
	if (tdm_event_class(event) != TDM_EVENT_SPAN) {
		tdm_error("%p: is not a span event\n", event);
		return;
	}

	if (!tdm_object_is_span(event->ev_obj.span)) {
		tdm_error("%p: object is not a span\n", event->ev_obj.span);
		return;
	}

	tdm_object_put(event->ev_obj.span);

	/* clean up per-event data */
}

tdm_status_t tdm_span_queue_event(struct tdm_span *span, const tdm_span_event_t ev_type, const void *data)
{
	struct tdm_event *event = NULL;
	tdm_status_t retval;

	if (!span || !tdm_object_is_span(span))
		return TDM_STATUS_GENERR;

	event = tdm_event_new(TDM_EVENT_SPAN);
	if (!event) {
		tdm_error("%s: failed to allocate new event %d\n", tdm_object_name(span), ev_type);
		return TDM_STATUS_GENERR;
	}

	event->ev_obj.span = tdm_object_get(span);
	event->ev_destroy  = &tdm_span_destroy_event;

	/* add per-event data? */
	switch (ev_type) {
	case TDM_SPAN_EVENT_SPAN_READY:
	case TDM_SPAN_EVENT_SPAN_STOPPED:
		break;
	default:
		if (data)
			event->ev_data.custom = data;
	}

	retval = tdm_event_queue(event);
	if (retval != TDM_STATUS_SUCCESS) {
		tdm_error("%s: failed to enqueue new event %d\n", tdm_object_name(span), ev_type);
		goto error;
	}

	return TDM_STATUS_SUCCESS;

error:
	tdm_span_destroy_event(event);
	return retval;
}


/*
 * dump device tree to stdout
 */
tdm_status_t tdm_debug_print_devtree(void)
{
	struct dlist_entry *m_entry, *d_entry, *p_entry;
	char prefix[20] = { 0 };

#define BRANCH_DELIM(p)	(dlist_is_last(p) ? "\\_" : "+-")
#define LINE_DELIM(p)	(dlist_is_last(p) ? " "  : "|")

	dlist_for_each(&tdm_module_list, m_entry) {
		struct tdm_module *mod = dlist_entry(m_entry, struct tdm_module, module_list);

		if (!tdm_object_is_module(mod)) {
			tdm_error("%p: object in module list is not a module", mod);
			break;
		}

		if (!mod->nr_devices)	/* no devices registered, skip */
			continue;

		printf("%s module \"%s\"\n", BRANCH_DELIM(m_entry), tdm_object_name(mod));
		printf("%s\tDevices:    %d\n", LINE_DELIM(m_entry), mod->nr_devices);
		printf("%s\n", LINE_DELIM(m_entry));

		/* walk module's device list */
		dlist_for_each(&mod->devices, d_entry) {
			struct tdm_device *dev = dlist_entry(d_entry, struct tdm_device, device_list);

			if (!tdm_object_is_device(dev)) {
				tdm_error("%p: object in device list is not a device", dev);
				break;
			}

			sprintf(prefix, "%s", LINE_DELIM(m_entry));
			printf("%s\t%s device \"%s\"\n", prefix, BRANCH_DELIM(d_entry), tdm_object_name(dev));

			sprintf(prefix, "%s\t%s", LINE_DELIM(m_entry), LINE_DELIM(d_entry));
			printf("%s\tPorts:    %d\n", prefix, dev->nr_ports);

			printf("%s\n", prefix);

			/* walk device's port list */
			dlist_for_each(&dev->port_list, p_entry) {
				struct tdm_port *port = dlist_entry(p_entry, struct tdm_port, port_list);

				if (!tdm_object_is_port(port)) {
					tdm_error("%p: object in port list is not a port", port);
					break;
				}

				sprintf(prefix, "%s\t%s", LINE_DELIM(m_entry), LINE_DELIM(d_entry));
				printf("%s\t%s port \"%s\"\n", prefix, BRANCH_DELIM(p_entry), tdm_object_name(port));

				sprintf(prefix, "%s\t%s\t%s", LINE_DELIM(m_entry), LINE_DELIM(d_entry), LINE_DELIM(p_entry));
				printf("%s\tChannels:    %d\n", prefix, port->nr_channels);

				if (!dlist_is_last(p_entry))
					printf("%s\n", prefix);
			}
		}
	}
#undef BRANCH_DELIM
#undef LINE_DELIM

	return TDM_STATUS_SUCCESS;
}


tdm_span_state_t tdm_span_get_state(struct tdm_span *span)
{
	if (!span || !tdm_object_is_span(span))
		return TDM_SPAN_STATE_INVALID;

	return span->state;
}

tdm_status_t tdm_span_start(struct tdm_span *span)
{
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!span || !tdm_object_is_span(span) || !span->ops)
		return TDM_STATUS_GENERR;

	/*
	 * Call (signalling) module's span start handler
	 */
	if (span->ops->span_start) {
		retval = span->ops->span_start(span);
		if (retval != TDM_STATUS_SUCCESS) {
			tdm_error("%s: error starting span: %s\n", tdm_object_name(span), tdm_errstr(ret));
			goto out;
		}
	}

	/* common span start */

#ifdef __TODO__
	pthread_create();

	/* TODO: span thread should set state to up/active and send SPAN_READY event */
#endif
out:
	return retval;
}

/*
 * possible change:
 *
 * typedef enum {
 * 	TDM_SHUTDOWN_GRACEFUL = 0,  // wait for all calls to complete (may block for a long time...)
 *	TDM_SHUTDOWN_IMMEDIATE      // kill all calls, shutdown span
 * } tdm_shutdown_type;
 *
 * #define TDM_SHUTDOWN_DEFAULT	TDM_SHUTDOWN_IMMEDIATE	// default: immediate
 *
 * tdm_span_stop(struct tdm_span *span, tdm_shutdown_type how)
 */

tdm_status_t tdm_span_stop(struct tdm_span *span)
{
	tdm_status_t retval = TDM_STATUS_SUCCESS;

	if (!span || !tdm_object_is_span(span) || !span->ops)
		return TDM_STATUS_GENERR;
#if __TODO__
	/*
	 * 1. put span into non-active state (= not accepting new calls)
	 *
	 * 2. for immediate:
	 *	- kill all active calls
	 *	- invoke callbacks
	 *	- mark span down/stopped
	 *	- send SPAN_STOPPED span event to user
	 *
	 *   for graceful:
	 *	- schedule callback for "all calls have ended"
	 *	- in callback:
	 *	- invoke span callbacks
	 *	- mark span down/stopped
	 *	- send SPAN_STOPPED span event to user
	 *
	 */
#endif

	/*
	 * Call (signalling) module's span stop handler
	 */
	if (span->ops->span_stop) {
		retval = span->ops->span_stop(span);
		if (retval != TDM_STATUS_SUCCESS) {
			tdm_error("%s: error stopping span: %s\n", tdm_object_name(span), tdm_errstr(ret));
			goto out;
		}
	}

	/* common span stop */
#ifdef __TODO__
	pthread_kill();
#endif

out:
	return retval;
}
